﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security;
using System.Security.Cryptography;
using System.Numerics;

namespace Paper
{
    namespace Token
    {

        public class BigIntegerUtils
        {

            public static BigInteger decodeBigInteger(String name, String str)
            {
                try
                {
                    return BigInteger.Parse(str);
                }
                catch (Exception e)
                {
                    throw new ArgumentOutOfRangeException(name, "Valore " + name + " non numerico");
                }
            }

            public static BigInteger fromBuffer(byte[] buff)
            {
                byte[] tmp = new byte[buff.Length];
                Buffer.BlockCopy(buff, 0, tmp, 0, buff.Length);
                Array.Reverse(tmp);   //java e' BIG endian, .net LITTLE endian
                return new BigInteger(tmp);
            }
        }

        public class RSA
        {

	    /* implementazione calcolo RSA con BigInteger */

            public static BigInteger encrypt(BigInteger m, BigInteger e, BigInteger n, String mask)
            {
                //equivalente java: BigInteger c = new BigInteger("0");
                BigInteger c = new BigInteger(0);

                //equivalente java: int nbitlength = n.NBitSize;
                int nbitlength = (n.ToByteArray().Length - 1) * 8;

                //equivalente java: BigInteger bitmask = (new BigInteger(mask)).pow(n.bitLength() - 1).subtract(new BigInteger("1"));
                BigInteger bitmask = new BigInteger(Int32.Parse(mask));
                bitmask = BigInteger.Pow(bitmask, nbitlength - 1);
                bitmask = BigInteger.Subtract(bitmask, new BigInteger(1));

                //equivalente java: BigInteger m2 = m;
                BigInteger m2 = m;

                int i = 0;

                BigInteger tmp;

                //equivalente java: while (m2.compareTo(bitmask) == 1)
                while (m2 > bitmask)
                {
                    //equivalente java: c = m2.and(bitmask).modPow(e, n).shiftLeft(i * n.bitLength()).or(c);
                    tmp = m2;
                    tmp &= bitmask;
                    tmp = BigInteger.ModPow(tmp, e, n);
                    tmp <<= (i * nbitlength);
                    tmp |= c;
                    c = tmp;

                    //equivalente java: m2 = m2.shiftRight(n.bitLength() - 1);
                    m2 >>= nbitlength - 1;

                    i = i + 1;
                }

                //equivalente java: c = m2.modPow(e, n).shiftLeft(i * n.bitLength()).or(c);
                tmp = m2;
                tmp = BigInteger.ModPow(tmp, e, n);
                tmp <<= (i * nbitlength);
                tmp |= c;
                c = tmp;

                return c;
            }
        }

        class TokenGenerator
        {

            public String buildTokenInfo(int abi, int filiale, String matricola, String profilo, DateTime dataOra)
            {
                if (abi > 9999)
                    throw new ArgumentOutOfRangeException("abi", "ABI " + abi + " non corretto, maggiore di 9999");

                if (filiale > 9999)
                    throw new ArgumentOutOfRangeException("filiale", "Codice filiale " + filiale + " non corretto, maggiore di 9999");

                if (matricola.Length != 5)
                    throw new ArgumentOutOfRangeException("matricola", "La matricola non corretta " + matricola + ", deve essere un codice di 4 cifre");

                if (profilo != "DP" && profilo != "SC")
                    throw new ArgumentOutOfRangeException("profilo", "Profilo " + profilo + " non corretto, deve essere SC o DP");

                return String.Format("{0,4:D4}", filiale) + ";" + String.Format("{0,4:D4}", abi) + ";" + matricola + ";" + profilo + ";" + dataOra.ToString("yyyy-MM-dd") + ";" + dataOra.ToString("HH:mm:ss");
            }

            private static String calcHash(String tokenInfo)
            {
                byte[] hashBytes = new SHA1CryptoServiceProvider().ComputeHash(System.Text.Encoding.Default.GetBytes(tokenInfo));
                string hashStr = System.Convert.ToBase64String(hashBytes, Base64FormattingOptions.None);
                return hashStr;
            }

            public String buildTokenInfoWithHash(int abi, int filiale, String matricola, String profilo, DateTime data)
            {
                String tokenInfo = buildTokenInfo(abi, filiale, matricola, profilo, data);

                return buildTokenInfoWithHash(tokenInfo);
            }

            public String buildTokenInfoWithHash(String tokenInfo)
            {
                return tokenInfo + ";" + calcHash(tokenInfo);
            }

            public String cryptToken(String token, BigInteger e, BigInteger n)
            {
                String tokenWithHash = token += ";" + calcHash(token);

                return cryptTokenWithHash(tokenWithHash, e, n);
            }

            public String cryptTokenWithHash(String tokenWithHash, BigInteger e, BigInteger n)
            {
                byte[] byteToCrypt = System.Text.Encoding.Default.GetBytes(tokenWithHash);

                BigInteger intToCrypt = BigIntegerUtils.fromBuffer(byteToCrypt);

                BigInteger crypted = RSA.encrypt(intToCrypt, e, n, "2");

                return crypted.ToString();
            }

            public String generate(int abi, int filiale, String matricola, String profilo, DateTime dataOra, BigInteger e, BigInteger n)
            {
                String tokenWithHash = buildTokenInfoWithHash(abi, filiale, matricola, profilo, dataOra);

                return cryptTokenWithHash(tokenWithHash, e, n);
            }

            public String generate(int abi, int filiale, String matricola, String profilo, DateTime dataOra, String strE, String strN)
            {
                String tokenWithHash = buildTokenInfoWithHash(abi, filiale, matricola, profilo, dataOra);

                BigInteger e = BigIntegerUtils.decodeBigInteger("e", strE);
                BigInteger n = BigIntegerUtils.decodeBigInteger("n", strN);

                return cryptTokenWithHash(tokenWithHash, e, n);
            }

            public String generate(String tokenInfo, BigInteger e, BigInteger n)
            {
                String tokenWithHash = buildTokenInfoWithHash(tokenInfo);

                return cryptTokenWithHash(tokenWithHash, e, n);
            }

            public String generate(String tokenInfo, String strE, String strN)
            {
                String tokenWithHash = buildTokenInfoWithHash(tokenInfo);

                BigInteger e = BigIntegerUtils.decodeBigInteger("e", strE);
                BigInteger n = BigIntegerUtils.decodeBigInteger("n", strN);

                return cryptTokenWithHash(tokenWithHash, e, n);
            }
        }
    }
}

